structure Check :>
sig

    exception TypeError of string

    (* Given an expression e, 
       return tau if there is a tau such that . |- e : tau, or
       raise TypeError otherwise *)
    val check : Syntax.exp -> Syntax.typ

end  = 
struct

    open Syntax
    
    type context = (name * typ) list
        
    exception TypeError of string

    fun lookup (gamma : context, v : name) : typ = 
	case List.find (fn (v', t) => v = v') gamma of
            NONE => raise TypeError ("unbound variable " ^ nameToString v)
          | SOME (_, t) => t

    (* Given a context gamma and expression e, 
       return tau if there is a tau such that gamma |- e : tau, or
       raise TypeError otherwise *)
    fun checkOpen (g : context, e : exp) : typ  = 
        case freshenTop e of 
            Var v => lookup (g,v)
          | Lam (t,(x,e)) => Arrow (t,checkOpen ( (x,t) :: g, e))
          | App (e1,e2) => 
                (case checkOpen ( g, e1) of
                     Arrow (t,t') => let val _ = checkWithExpected ( g, e2, t) in t' end
                   | _ => raise TypeError "function position not of function type")
          | Let (e1, t, (x,e2)) => checkOpen ( ((x,checkWithExpected ( g, e1, t)) :: g), e2 )
          | Zero => Nat
          | Succ e => checkWithExpected(g,e,Nat)
          | NatCase (e, e0, (x,e1)) =>
                            let
                                val _ = checkWithExpected( g,e,Nat)
                                val t = checkOpen( g,e0)
                                val _ = checkWithExpected( (x,Nat) :: g, e1, t)
                            in
                                t
                            end
          | Nil => List
          | Cons(e1, e2) => let val _ = checkWithExpected( g,e1,Nat)
                                val _ = checkWithExpected( g,e2,List)
                            in 
                                List
                            end
          | ListCase (e, e0, (x,y,e1)) =>
                            let
                                val _ = checkWithExpected( g,e,List)
                                val t = checkOpen( g,e0)
                                val _ = checkWithExpected( (x,Nat) :: (y,List) :: g, e1, t)
                            in
                                t
                            end
          | Fix(t,(x,e)) => checkWithExpected( (x,t) :: g, e, t)
          | Subst(e,s) => raise Fail "case for substitution unimplemented"

    and checkWithExpected(g : context, e : exp, t : typ) : typ =
        if checkOpen(g,e) = t then t else raise TypeError "expression doesn't have expected type"
                     
    fun check (e) = checkOpen ([], e)
                      
end
